# measpy/ni.py
#
# -------------------------------------------------
# Data acquisition with National Instrument devices
# -------------------------------------------------
#
# Part of measpy package for signal acquisition and processing
# (c) OD - 2021 - 2023
# https://github.com/odoare/measpy

from datetime import datetime
from time import sleep

import nidaqmx
from nidaqmx import stream_readers
import nidaqmx.constants as niconst
from threading import Event
import numpy as np
import h5py

from ._tools import siglist_to_array, t_min, _add_N_data, H5file_valid
from .signal import Signal, SignalType

def _n_to_ain(n):
    return 'ai'+str(n-1)
def _n_to_aon(n):
    return 'ao'+str(n-1)
def _n_to_don(n):
    return 'port'+str(n-1)

def ni_run_measurement(M, filename=None, duration="default", chunck_size = 0):
    """
    Runs a measurement defined in the object of
    the class measpy.measurement.Measurement given
    as argument.

    Run callback at refresh_rate with data acquired

    Once the data acquisition process is terminated,
    the measurement object given in argument contains
    a property in_sig consisting of a list of signals.

    :param M: The measurement object that defines the measurement properties
    :type M: measpy.measurement.Measurement
    :param filename: .h5 filename to direct write on disk, defaults to None
    :type filename: str or Path, optional
    :param duration: optional duration in second, take value in M is default, defaults to "default"
    :type duration: float, optional
    :return: Nothing, the measurement passed as argument is modified in place.

    """
    if filename is not None:
        if H5file_valid(filename):
            # M.to_hdf5(filename)
            M.create_hdf5(filename,chunck_size=chunck_size)

            def callback(buffer_in, n_values):
                _add_N_data(H5file["in_sig"], buffer_in, n_values)

            with h5py.File(filename, "r+") as H5file:
                try:
                    n_values, Nchannel = H5file["in_sig"].chunks
                except ValueError:
                    n_values = H5file["in_sig"].chunks[0]
                with ni_callback_measurement(M) as NI:
                    NI.set_callback(callback, n_values)
                    NI.run(duration=duration)
            M.load_h5data()
        else:
            if input("Write in file canceled, cancel measurement ? y/n\n") == "y":
                return
    else:
        samples = []

        def callback(buffer_in, n_values):
            nonlocal samples
            samples.extend(buffer_in.copy())

        n_values = min(2**14,int(0.1*M.dur*M.fs))
        with ni_callback_measurement(M) as NI:
            NI.set_callback(callback, n_values)
            NI.run(duration=duration)
        if isinstance(M.in_sig, Signal):
            M.in_sig.raw = np.fromiter(
                samples, np.dtype((float, (len(M.in_map),))), count=len(samples)
            ).squeeze()
            M.in_sig.t0 = NI.tmin
        else:
            if len(M.in_map) == 1:
                M.in_sig[0].raw = np.fromiter(
                    samples,
                    np.dtype(float),
                    count=len(samples),
                )
                M.in_sig[0].t0 = NI.tmin
            else:
                data = np.fromiter(
                    samples,
                    np.dtype((float, (len(M.in_map),))),
                    count=len(samples),
                )
                for i, s in enumerate(M.in_sig):
                    s.raw = data[:, i]
                    s.t0 = NI.tmin
    print("done")


class ni_callback_measurement:
    """
    Measurment using a callback function called when specified number of sample is written from the device to the buffer.
    Using nidaqmx.Task.register_every_n_samples_acquired_into_buffer_event
    """

    def __init__(self, M):
        #Parameters setup
        system = nidaqmx.system.System.local()
        self.callback_set = False
        self.Nchannel = len(M.in_map)

        if isinstance(M.in_sig, list) and len(M.in_sig) != self.Nchannel:
            raise ValueError(
                f"in_sig property of measurement must be a multichannel signal or a list of {len(M.in_map)} single channel signals"
            )

        self.in_multichannel = isinstance(M.in_sig, Signal)

        if isinstance(M.out_sig, list) and len(M.out_sig) != len(M.out_map):
            raise ValueError(
                f"out_sig property of measurement must be a multichannel signal or a list of {len(M.out_map)} single channel signals"
            )

        out_multichannel = isinstance(M.out_sig, Signal)

        if M.device_type != "ni":
            print("Warning: deviceType != 'ni'. Changing to 'ni'.")
            M.device_type = "ni"
        if M.in_device == "":
            print(
                "Warning: no output device specified, changing to "
                + system.devices[0].name
            )
            M.in_device = system.devices[0].name

        if hasattr(M, "in_range"):
            inr = M.in_range is not None
        else:
            inr = False
        if not (inr):
            val = nidaqmx.system.device.Device(M.in_device).ai_voltage_rngs[-1]
            print(
                "Warning: no input range specified, changing to the max value of "
                + M.in_device
                + " -> "
                + str(val)
            )
            M.in_range = list(val for b in M.in_map)

        if hasattr(M, "out_sig") and M.out_sig is not None:
            if out_multichannel:
                self.outx = siglist_to_array(M.out_sig.unpack())
            else:
                self.outx = siglist_to_array(M.out_sig)
            self.tmin = t_min(M.out_sig)
            if M.out_device == "":
                print(
                    "Warning: no output device specified, changing to "
                    + system.devices[0].name
                )
                M.out_device = system.devices[0].name
            if hasattr(M, "out_range"):
                if M.out_range is None:
                    outr = False
                else:
                    outr = True
            else:
                outr = False
            if not (outr):
                val = nidaqmx.system.device.Device(M.out_device).ao_voltage_rngs[
                    -1
                ]
                print(
                    "Warning: no output range specified, changing to the max value of "
                    + M.out_device
                    + " -> "
                    + str(val)
                )
                M.out_range = list(val for b in M.out_map)
        else:
            self.tmin = 0
        self.M = M

    def __enter__(self):
        #Open and set up NI Tasks
        try:
            # Set up the read tasks
            if self.M.in_sig is not None:
                self.intask = nidaqmx.Task(new_task_name="in")  # read task
                for i, n in enumerate(self.M.in_map):
                    print(_n_to_ain(n))
                    self.intask.ai_channels.add_ai_voltage_chan(
                        physical_channel=self.M.in_device + "/" + _n_to_ain(n),
                        terminal_config=niconst.TerminalConfiguration.DEFAULT,
                        min_val=-self.M.in_range[i],
                        max_val=self.M.in_range[i],
                        units=niconst.VoltageUnits.VOLTS,
                    )

                self.intask.timing.cfg_samp_clk_timing(
                    rate=self.M.fs,
                    sample_mode=niconst.AcquisitionType.CONTINUOUS,
                )

                for i, iepeval in enumerate(self.M.in_iepe):
                    if iepeval:
                        self.intask.ai_channels[i].ai_excit_val = 0.002
                        self.intask.ai_channels[
                            i
                        ].ai_coupling = niconst.Coupling.AC

            self.reader = stream_readers.AnalogMultiChannelReader(
                self.intask.in_stream
            )

            # Set up the write tasks
            if self.M.out_sig is not None:
                self.nsamps = int(round(self.M.dur * self.M.fs))
                self.outtask = nidaqmx.Task(new_task_name="out")  # write task

                if self.M.out_sig[0].type == SignalType.ANALOG:
                    # Set up the write tasks, use the sample clock of the Analog input if possible
                    for i, n in enumerate(self.M.out_map):
                        self.outtask.ao_channels.add_ao_voltage_chan(
                            physical_channel=self.M.out_device + "/" + _n_to_aon(n),
                            min_val=-self.M.out_range[i],
                            max_val=self.M.out_range[i],
                            units=niconst.VoltageUnits.VOLTS,
                        )
                elif self.M.out_sig[0].type == SignalType.DIGITAL:
                    for i, n in enumerate(self.M.out_map):
                        self.outtask.do_channels.add_do_chan(
                            lines=self.M.out_device + "/" + _n_to_don(n),
                            line_grouping=niconst.LineGrouping.CHAN_FOR_ALL_LINES
                        )

                if self.M.in_device.startswith("myDAQ") and self.M.out_sig.type == SignalType.ANALOG:
                    # If the device is a myDAQ card, we keep most default values
                    # The myDAQ devices are set up separately because
                    # there is no error messages when setting up properties
                    # that are not supported, and the acquisition then fails
                    print("This is a NI my DAQ device")
                    self.outtask.timing.cfg_samp_clk_timing(
                        rate=self.M.fs,
                        sample_mode=niconst.AcquisitionType.CONTINUOUS,
                        samps_per_chan=self.nsamps,
                    )
                else:
                    try:
                        # We first try to use analog input sample clock as output clock
                        self.outtask.timing.cfg_samp_clk_timing(
                            rate=self.M.fs,
                            source="/"
                            + self.M.in_device
                            + "/ai/SampleClock",  # "OnboardClock",
                            sample_mode=niconst.AcquisitionType.CONTINUOUS,
                            samps_per_chan=self.nsamps,
                        )
                        print(
                            "Use of /"
                            + self.M.in_device
                            + "/ai/SampleClock as output clock : success !"
                        )
                    except:
                        # If it fails, use defaults
                        # Then the in/out are not synchronized
                        # There is hence the possibility to use one analog input
                        # to do the in/out sync (io_sync=input channel number)
                        print(
                            'Choosing "'
                            + "/"
                            + self.M.in_device
                            + '/ai/SampleClock" as clock source causes trouble, let\'s try "OnboardClock" '
                        )
                        self.outtask.timing.cfg_samp_clk_timing(
                            rate=self.M.fs,
                            sample_mode=niconst.AcquisitionType.CONTINUOUS,
                            samps_per_chan=self.nsamps,
                        )

                print("values: ",self.outx[:, 0])
                if len(self.M.out_map) == 1:
                    if self.M.out_sig[0].type == SignalType.DIGITAL:
                        self.outtask.write(list(map(int,map(round,self.outx[:, 0]))), auto_start=False) 
                    else:
                        self.outtask.write(self.outx[:, 0], auto_start=False)
                else:
                    # If there are more than one output channel,
                    # the outx.T array argument produces an error.
                    # Temporary dirty fix consists of converting
                    # the array to a list.
                    # TODO: Find better solution
                    self.outtask.write((self.outx.T).tolist(), auto_start=False)
        except Exception as e:
            self.__exit__(type(e), e.args, e)
            raise e
        return self

    def __exit__(self, exc_type, exc_value, tb):
        #Close all tasks
        if self.M.in_sig != None:
            self.intask.close()
        if self.M.out_sig != None:
            self.outtask.close()
        print("ni tasks closed")

    def run(self, stop=Event(), duration="default"):
        """
        Run the measurment
        :param stop: Trigger to stop the measurment, defaults to Event()
        :type stop: threading.Event, optional
        :param duration: Duration of measurment in seconds if default it use the duration in M, defaults to "default"
        :type duration: float, optional
        :return: Nothing

        """
        if not self.callback_set:
            print("Cancelled: there is no callback set")
            return
        self.now = datetime.now()
        self.M.date = self.now.strftime("%Y-%m-%d")
        self.M.time = self.now.strftime("%H:%M:%S")
        if not self.in_multichannel:
            for s in self.M.in_sig:
                s.t0 = self.tmin
        else:
            self.M.in_sig.t0 = self.tmin
        if duration:
            if duration == "default":
                numdesiredsamples = int(round(self.M.fs * self.M.dur))
            else:
                numdesiredsamples = int(round(self.M.fs * duration))
            numBuffersToCapture = int(np.ceil(numdesiredsamples / self.n_values))
        else:
            numBuffersToCapture = float("inf")
        self.buffer_captured = 0
        time_to_fill_buffer = self.n_values / self.M.fs
        if self.M.out_sig != None:
            self.outtask.start()  # Start the write task first, waiting for the analog input sample clock
        if self.M.in_sig != None:
            self.intask.start()
        try:
            print("Start measurement")
            while (
                not stop.is_set() and self.buffer_captured < numBuffersToCapture
            ):
                sleep(time_to_fill_buffer)
        except KeyboardInterrupt:
            pass
        self.stop()

    def stop(self):
        #stop all task
        if self.M.in_sig != None:
            self.intask.stop()
        if self.M.out_sig != None:
            self.outtask.stop()
        print("ni tasks stopped")

    def set_callback(self, callback_method, n_values):
        """
        Create the buffer containing n_values and set the callback that read and use the data using
        the custom 'callback_method', it is called every 'n_values' written into buffer.
        :param callback_method: Method with 2 arguments, the buffer and buffer lenght
        :type callback_method: callable
        :param n_values: Number of datapoint read each call
        :type n_values: int
        :return: Nothing

        """
        self.n_values = n_values
        self.buffer_captured = 0
        if self.Nchannel > 1:
            buffer_in = np.zeros((self.Nchannel, self.n_values), dtype="f8")

            def callback(task_handle, event_type, n_values, callback_data):
                self.reader.read_many_sample(
                    buffer_in, n_values, timeout=niconst.WAIT_INFINITELY
                )
                callback_method(buffer_in.T, n_values)
                self.buffer_captured += 1
                return 0

        else:
            buffer_in = np.zeros((1, self.n_values), dtype="f8")

            def callback(task_handle, event_type, n_values, callback_data):
                self.reader.read_many_sample(
                    buffer_in, n_values, timeout=niconst.WAIT_INFINITELY
                )
                callback_method(buffer_in[0, :], n_values)
                self.buffer_captured += 1
                return 0

        self.intask.register_every_n_samples_acquired_into_buffer_event(
            self.n_values, callback
        )
        self.intask.in_stream.input_buf_size = int(10 * self.n_values)
        self.callback_set = True
        print("callback set")

    def reset_callback(self, callback_method, n_values):
        self.intask.register_every_n_samples_acquired_into_buffer_event(
            self.n_values, None
        )
        self.set_callback(callback_method, n_values)

def ni_run_synced_measurement(M,in_chan=0,out_chan=0):
    """
    Before running a measurement, added_time second of silence
    is added at the begining and end of the selected output channel.
    The measurement is then run, and the time lag between
    a selected acquired signal and the output signal is computed
    from cross-correlation calculation.
    All the acquired signals are then re-synced from the time lag value.

    :param M: The measurement object
    :type M: measpy.measurement.Measurement
    :param out_chan: The selected output channel for synchronization. It is the index of the selected output signal in the list ``M.out_sig``
    :type out_chan: int
    :param in_chan: The selected input channel for synchronization. It is the index of the selected input signal in the list ``M.in_sig``
    :type in_chan: int
    :param added_time: Duration of silence added before and after the selected output signal
    :type added_time: float
    :return: Measured delay between i/o sync channels
    :rtype: float
    """
    M.sync_prepare(out_chan=out_chan)
    ni_run_measurement(M)
    d = M.sync_render(in_chan=in_chan,out_chan=out_chan)
    return d

def ni_get_devices():
    """
    Get the list of NI devices present in the system

    :returns: A list of devices object
    """
    system = nidaqmx.system.System.local()
    print(system.devices)
    return system.devices



# measpy/measurement.py
#
# ------------------------------------------------------
# Measuremnt class definition for data acquisition tasks
# ------------------------------------------------------
#
# Part of measpy package for signal acquisition and processing
# (c) OD - 2021 - 2023
# https://github.com/odoare/measpy

from os.path import isfile
from copy import copy,deepcopy
import csv
import os

from functools import partial
import numpy as np
import h5py

from .signal import Signal

from ._tools import (csv_to_dict, 
                     convl, 
                     convl1,  
                     calc_dur_siglist,
                     h5file_write_from_queue)

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions


class Measurement:
    # ---------------------------
    def __init__(self, **params):
        # Check out_sig contents
        if 'out_sig' in params:

            non=params['out_sig'] is not None

            # print(type(params['in_sig']))

            if non & (not isinstance(params['out_sig'],(Signal,list))):
                raise TypeError("out_sig must but be a Signal, a list of measpy.signal.Signal or None")

            if isinstance(params['out_sig'],Signal):
                self.out_sig = params['out_sig']
            elif isinstance(params['out_sig'],list):
                if len(params['out_sig']) != len(params['out_map']):
                    raise ValueError('Measurement out_map has not the same number of values than the number of signals in out_sig list')
                #print(list((type(s)==Signal for s in params['out_sig'])))
                if all((isinstance(s,Signal) for s in params['out_sig'])):
                    # print('These are all signals')
                    if all(s.fs==params['out_sig'][0].fs for s in params['out_sig']):
                        # print('Same fs for all signals')
                        self.out_sig = params['out_sig']
                    else:
                        raise ValueError("Signals in out_sig list have different sampling frequencies")                     
            else:
                raise TypeError("Some elements of out_sig list are not Signals")
        else:
            self.out_sig = None

        # Check in_sig contents
        if 'in_sig' in params:

            non=params['in_sig'] is not None
            
            if non & (not isinstance(params['in_sig'],(Signal,list))):
                raise TypeError("in_sig must but be a Signal or a list of measpy.signal.Signal or None")
            
            if isinstance(params['in_sig'],Signal):
                self.in_sig = params['in_sig']
            elif isinstance(params['in_sig'],list):
                if len(params['in_sig']) != len(params['in_sig']):
                    raise ValueError('Measurement in_map has not the same number of values than the number of signals in in_sig list')
                if all((isinstance(s,Signal) for s in params['in_sig'])):
                    # print('These are all signals')
                    if all(s.fs==params['in_sig'][0].fs for s in params['in_sig']):
                        # print('Same fs for all signals')
                        self.in_sig = params['in_sig']
                    else:
                        raise ValueError("Signals in in_sig list have different sampling frequencies")                     
                else:
                    raise TypeError("Some elements of in_sig list are not Signals")
        else:
            self.in_sig = None

        #Check sampling frequencies
        if isinstance(self.out_sig,type(None)):
            if isinstance(self.in_sig,type(None)):
                #raise Exception("This is a task with no input nor output ?")
                print("This is a task with no input nor output ?")
            else:
                # print("This is a task with no output.")
                if 'fs' in params:
                    if params['fs']!=self.in_sig[0].fs:
                        print('Selected sampling frequency '+str(params['fs'])+'Hz is different to that given in signals in in_sig list')
                        print('Sampling frequencies of all input signals are set to the selected value: '+str(params['fs'])+'Hz.')
                        self.fs = params['fs']
                        for s in self.in_sig:
                            s.fs = params['fs']
                    else:
                        self.fs = params['fs']
                else:
                    self.fs = self.in_sig[0].fs
                    print('Task frequency is: ', str(self.fs) )
        else:
            if 'fs' in params:
                if params['fs']!=self.out_sig[0].fs:
                    print('Selected sampling frequency '+str(params['fs'])+'Hz is different to that given in signals in out_sig list')
                    print("Task's sampling frequency is set to the selected value in outpu signals: "+str(self.out_sig[0].fs)+"Hz.")
            self.fs = self.out_sig[0].fs
            if type(self.in_sig)==type(None):
                print ("This is a task with no input.")
            else:
                if self.fs!=self.in_sig[0].fs:
                    print('Selected sampling frequency '+str(self.fs)+'Hz is different to that given in signals on in_sig list')
                    print('Sampling frequencies of all input signals are set to the selected value: '+str(self.fs)+'Hz.')
                    for s in self.in_sig:
                        s.fs = self.fs

        # Check list lengths
        if type(self.out_sig)!=type(None):
            if 'out_map' in params:
                if len(params['out_map'])!=len(self.out_sig):
                    if len(self.out_sig)!=1:
                        raise Exception('Lengths of out_map and out_sig do not correspond.')
                    else:
                        print(f"out_sig contains one signal whereas len(out_map)={len(params['out_map'])}, the signal will be multichannel")
                self.out_map = params['out_map']
            else:
                self.out_map = list(range(1,len(self.out_sig)+1))
                print("out_map not given, it is set to default value of: "+str(self.out_map))
        if type(self.in_sig)!=type(None):
            if 'in_map' in params:
                if len(params['in_map'])!=len(self.in_sig):
                    if len(self.in_sig)!=1:
                        raise Exception('Lengths of in_map and in_sig do not correspond.')
                    else:
                        print(f"in_sig contains one signal whereas len(in_map)={len(params['in_map'])}, the signal will be multichannel")    
                self.in_map = params['in_map']
            else:
                self.in_map = list(range(1,len(self.in_sig)+1))
                print("in_map not given, it is set to default value of: "+str(self.in_map))
            if 'in_threshold' in params:
                if len(params['in_threshold'])!=len(self.in_map):
                    raise Exception('Lengths of in_threshold and in_map do not correspond.')
                self.in_threshold = params['in_threshold']

        self.in_device = params.setdefault("in_device",'')
        self.out_device = params.setdefault("out_device",'')

        # Check durations
        if 'dur' in params:
            if type(self.out_sig)==type(None):
                self.dur = params['dur']
            else:
                dursigs = calc_dur_siglist(self.out_sig)
                if params['dur']!=dursigs:
                    print('Selected duration is different thant duration of combined output signals.')
                    print('It is changed to match.')
                self.dur = dursigs
        else:
            if type(self.out_sig)==type(None):
                #raise Exception('No duration nor out_sig given. Impossible to determine task duration')
                print('No duration nor out_sig given. Impossible to determine task duration')
            else:
                self.dur = calc_dur_siglist(self.out_sig)
                print("Duration of the task set to: "+str(self.dur)+" s.")

        if 'device_type' not in params:
            self.device_type = ''
            print('No device_type given, it is set to empty string and will be updated when performing the task.')
        else:
            self.device_type = params['device_type']

        # Fix specific properties
        if self.device_type=='pico':
            self.in_range = params.setdefault("in_range",list('10V' for b in self.in_map))
            self.upsampling_factor = params.setdefault("upsampling_factor",1)
            self.in_coupling = params.setdefault("in_coupling",list('dc' for b in self.in_map))
            self.sig_gen = params.setdefault("sig_gen",False)
            if self.sig_gen is not None:
                self.offset = params.setdefault("offset",0.0)
                self.wave = params.setdefault("wave",0)
                self.amp = params.setdefault("amp",1.0)
                self.freq_start = params.setdefault("freq_start",20)
                self.freq_stop = params.setdefault("freq_start",20_000)
                self.freq_change = params.setdefault("freq_change",10)
                self.freq_int = params.setdefault("freq_int",0.01)
                self.sweep_dir = params.setdefault("sweep_dir",0)
                self.sweep_number = params.setdefault("sweep_number",100)
        if self.device_type=='ni':
            self.in_range = params.setdefault("in_range",None)
            self.out_range = params.setdefault("out_range",None)
            self.in_iepe = params.setdefault("in_iepe",list(False for b in self.in_map))
        if type(self.out_sig)!=type(None):
            self.io_sync = params.setdefault('io_sync',0)
        elif 'io_sync' in params:
                print('No output signals given. io_sync param ignored')
        self.desc = params.setdefault('desc','No description')

    # -----------------
    def __repr__(self):
        out = "measpy.Measurement("
        out += "fs="+str(self.fs)
        out += ",\n dur="+str(self.dur)
        out += ",\n device_type='"+str(self.device_type)+"'"
        if self.in_sig!=None:
            out += ',\n in_device='+str(self.in_device)
            out += ',\n in_map='+str(self.in_map)
        try:
            out += ",\n date='"+self.date+"'"
            out += ",\n time='"+self.time+"'"
        except:
            pass
        if self.out_sig!=None:
            out += ",\n out_device="+str(self.out_device)
            out += ',\n out_map='+str(self.out_map)
            if isinstance(self.out_sig,list):
                out += ',\n out_sig=list of '+str(len(self.out_sig))+' measpy.signal.Signal'
            else:
                out += ',\n out_sig=multichannel measpy.signal.Signal'
            out += ",\n io_sync="+str(self.io_sync)
        if self.device_type=='pico':
            out += ",\n in_range="+str(self.in_range)
            out += ",\n upsampling_factor="+str(self.upsampling_factor)
            out += ",\n in_coupling="+str(self.in_coupling)
        if self.device_type=='ni':
            out += ",\n in_range="+str(self.in_range)
            out += ",\n out_range="+str(self.out_range)
            out += ",\n in_iepe="+str(self.in_iepe)            
        if isinstance(self.in_sig,list):
            out += ',\n in_sig=list of '+str(len(self.in_sig))+' measpy.signal.Signal'
        else:
            out += ',\n in_sig=multichannel measpy.signal.Signal'
        out +=")"
        
        return out
    
    # ------------------------------
    def _to_dict(self,withsig=True):
        """ Converts a Measurement object to a dict

            :param withsig: Optionnally removes the data arrays, defaults to True
            :type withsig: bool
            
        """
        mesu = copy(self.__dict__)
        if not(withsig):
            del mesu['in_sig']
            del mesu['out_sig']
        return mesu

    # -----------------------
    def to_dir(self,dirname):
        """ Stores the parameters and signals in a directory

            :param dirname: Name of the directory, a (1), (2)... is added to the name if directory exists
            :type dirname: str
            :return: Actual name to the saved folder (if name conflit is detected)
            :rtype: str                            
        """
        if os.path.exists(dirname):
            i = 1
            while os.path.exists(dirname+'('+str(i)+')'):
                i+=1
            dirname = dirname+'('+str(i)+')'
        os.mkdir(dirname)
        self._params_to_csv(dirname+"/params.csv")
        if isinstance(self.in_sig,list):
            for i,s in enumerate(self.in_sig):
                s.to_csvwav(dirname+"/in_sig_"+str(i))
        elif isinstance(self.in_sig,Signal):
            self.in_sig.to_csvwav(dirname+"/in_sig")
        if isinstance(self.out_sig,list):
            for i,s in enumerate(self.out_sig):
                s.to_csvwav(dirname+"/out_sig_"+str(i))
        elif isinstance(self.out_sig,Signal):
            self.out_sig.to_csvwav(dirname+"/out_sig")
        self._write_readme(dirname+"/README")
        return dirname

    def to_hdf5(self, filename):
        """
        Save Measurement in hdf5 file

        :param filename: hdf5 file
        :type filename: str

        """
        mesu = self._to_dict()
        in_sig = mesu.pop("in_sig",None)
        if not any([s.dur>0 for s in in_sig]):
            print("There is no data in this measurement, use 'create_hdf5' instead")
            return
        out_sig = mesu.pop("out_sig",None)
        with h5py.File(filename, "x") as H5file:
            for name, value in mesu.items():
                if value is not None:
                    H5file.attrs[name] = value
            if isinstance(in_sig,list):
                in_sig = Signal.pack(in_sig)
            if isinstance(in_sig,Signal):
                in_sig.to_hdf5(H5file, "in_sig")
                # self.h5save_data = in_sig.h5save_data
            if isinstance(out_sig,list):
                out_sig = Signal.pack(out_sig)
            if isinstance(out_sig,Signal):
                out_sig.to_hdf5(H5file, "out_sig")
        self.filename = filename

    def create_hdf5(self, filename, chunck_size=0, datatype=None, dbfs=None):
        mesu = self._to_dict()
        in_sig = mesu.pop("in_sig",None)
        out_sig = mesu.pop("out_sig",None)
        data_type = mesu.pop("data_type",datatype)
        with h5py.File(filename, "x") as H5file:
            for name, value in mesu.items():
                if value is not None:
                    H5file.attrs[name] = value
            if isinstance(out_sig,list):
                out_sig = Signal.pack(out_sig)
            if isinstance(out_sig,Signal):
                out_sig.to_hdf5(H5file, "out_sig", data_type)
            if isinstance(in_sig,list):
                in_sig = Signal.pack(in_sig)
            if isinstance(in_sig,Signal):
                if self.device_type=='pico':
                    # picoscope read data in increasing channel number order, remaping to match in_map order
                    Channel_map = np.argsort(np.argsort(self.in_map))
                else:
                    Channel_map = None
            self.filename = filename
            return in_sig.create_hdf5dataset(
                H5file,
                chunck_size=chunck_size,
                dataset_name="in_sig",
                data_type=data_type,
                Channel_map=Channel_map,
                dbfs=dbfs,
                )


    def load_h5data(self):
        if hasattr(self, "filename"):
            print(f"Loading data from {self.filename}")
            with h5py.File(self.filename, "r") as H5file:
                self._load_h5signal(H5file)
        else:
            print("No hdf5 file associated with this measurement")

    def _load_h5signal(self, H5file):
        try:
            self.in_sig = Signal.from_hdf5(H5file["in_sig"])
        except KeyError:
            self.in_sig = None
        try:
            self.out_sig = Signal.from_hdf5(H5file["out_sig"])
        except KeyError:
            self.out_sig = None

    # ------------------------
    @classmethod
    def from_dir(cls,dirname):
        """ Load a measurement object from a directory

            :param dirname: Name of the directory
            :type dirname: str                
        """

        task_dict = csv_to_dict(dirname+'/params.csv')
        self = cls._from_dict(task_dict)

        if 'in_map' in task_dict:
            if isfile(f'{dirname}/in_sig.wav'):
                print('in_sig is a multichannel signal')
                self.in_sig = Signal.from_csvwav(f'{dirname}/in_sig')
            else:
                print('in_sig is a list of signals')
                self.in_sig = list(Signal.from_csvwav(dirname+'/in_sig_'+str(i)) for i in range(len(task_dict['in_map'])) )
        if 'out_map' in task_dict:
            if isfile(f'{dirname}/out_sig.wav'):
                print('out_sig is a multichannel signal')
                self.out_sig = Signal.from_csvwav(f'{dirname}/out_sig')
            else:
                print('out_sig is a list of signals')
                self.out_sig = list(Signal.from_csvwav(dirname+'/out_sig_'+str(i)) for i in range(len(task_dict['out_map'])) )

        return self

    @classmethod
    def from_hdf5(cls,filename):
        with h5py.File(filename, "r") as H5file:
            task_dict = {}
            for key,val in H5file.attrs.items():
                task_dict[key] = val
            self = cls._from_dict(task_dict)
            self._load_h5signal(H5file)
        self.filename = filename
        return self

    @classmethod
    def _from_dict(cls, task_dict):
        self=cls()
        self.fs=convl1(float,task_dict['fs'])
        self.dur=convl1(float,task_dict['dur'])
        try:
            self.date=convl1(str,task_dict['date'])
            self.time=convl1(str,task_dict['time'])
        except:
            pass
        self.device_type=convl1(str,task_dict['device_type'])

        if 'in_map' in task_dict:
            self.in_map = convl(int,task_dict['in_map'])
            if 'in_device' in task_dict:
                self.in_device = convl1(str,task_dict['in_device'])
            else:
                self.in_device = None
        else:
            self.in_sig = None
        if 'out_map' in task_dict:
            self.out_map = convl(int,task_dict['out_map'])
            if 'out_device' in task_dict:
                self.out_device = convl1(str,task_dict['out_device'])
            else:
                self.out_device = None
            if 'io_sync' in task_dict:
                self.io_sync = convl1(int,task_dict['io_sync'])
        else:
            self.out_sig = None

        # Picoscope specifics
        if self.device_type == 'pico':
            try:
                self.in_range = convl(str,task_dict['in_range'])
            except:
                pass
            try:
                self.in_coupling = convl(str,task_dict['in_coupling'])
            except:
                pass
            try:
                self.upsampling_factor = convl1(int,task_dict['upsampling_factor'])
            except:
                pass
            try:
                self.sig_gen = convl1(bool,task_dict['sig_gen'])
            except:
                pass
            try:
                self.offset = convl1(float,task_dict['offset'])
            except:
                pass
            try:
                self.wave = convl1(int,task_dict['wave'])
            except:
                pass
            try:
                self.freq_start = convl1(float,task_dict['freq_start'])
            except:
                pass
            try:
                self.freq_stop = convl1(float,task_dict['freq_stop'])
            except:
                pass
            try:
                self.freq_change = convl1(float,task_dict['freq_change'])
            except:
                pass
            try:
                self.freq_int = convl1(float,task_dict['freq_int'])
            except:
                pass
            try:
                self.sweep_dir = convl1(int,task_dict['sweep_dir'])
            except:
                pass
            try:
                self.sweep_number = convl1(float,task_dict['sweep_number'])
            except:
                pass

        # NI specifics
        if self.device_type == 'ni':
            try:
                self.in_range = convl1(float,task_dict['in_range'])
            except:
                self.in_range = None
            try:
                self.out_range = convl1(float,task_dict['out_range'])
            except:
                self.out_range = None
            try:
                self.in_iepe = convl(bool,task_dict['in_iepe'])
            except:
                pass

        return self

    # --------------------------------
    def _params_to_csv(self,filename):
        """ Writes all the Measurement object parameters to a csv file """
        dd = self._to_dict(withsig=False)
        with open(filename, 'w', newline='') as file:
            writer = csv.writer(file)
            for key in dd:
                if type(dd[key])==list:
                    writer.writerow([key]+dd[key])
                else:
                    writer.writerow([key,str(dd[key])])

    # -------------------------------
    def _write_readme(self,filename):
        with open(filename, 'w') as f:
            f.write('Created with measpy version '+__version__)
            f.write('\n')
            f.write('https://github.com/odoare/measpy')

    def sync_prepare(self,out_chan=0,added_samples=None):
        """
        Prepare measurement for synchronization

        :param out_chan: The selected output channel for synchronization. It is the index of the selected output signal in the list ``M.out_sig``
        :type out_chan: int
        :param in_chan: The selected input channel for synchronization. It is the index of the selected input signal in the list ``M.in_sig``
        :type in_chan: int
        :param added_time: Duration of silence added before and after the selected output signal
        :type added_time: float
        """
        if type(added_samples)==type(None):
            asp = self.fs
        elif type(added_samples)==int:
            asp = added_samples
        else:
            raise TypeError("added_samples should an integer value")
        if isinstance(self.out_sig,Signal):
            self.out_sig = self.out_sig.add_silence(extras=(asp,asp)).delay(-asp/self.fs)
            self.dur = self.out_sig.dur
        else:
            osig = self.out_sig[out_chan].add_silence(extras=(asp,asp)).delay(-asp/self.fs)
            self.dur = osig.dur
            self.out_sig[out_chan]=osig

    def sync_render(self,out_chan=0,in_chan=0,added_samples=None):
        if type(added_samples)==type(None):
            asp = self.fs
        elif type(added_samples)==int:
            asp = added_samples
        else:
            raise TypeError("added_samples: Wrong type")
        d = self.in_sig[in_chan].timelag(self.out_sig[out_chan])
        ds = round(d*self.fs)
        # dt = 1/self.fs
        # print("delay: "+str(d)+"s")
        self.dur = self.dur-2*asp/self.fs
        if isinstance(self.out_sig,Signal):
            self.out_sig = self.out_sig.cut(pos=(asp,asp+round(self.dur*self.fs))).delay(asp/self.fs)
            self.in_sig = self.in_sig.cut(pos=(asp+ds+1,asp+ds+1+self.out_sig[out_chan].length))
            self.in_sig.t0 = self.out_sig.t0
        else:
            self.out_sig[out_chan] = self.out_sig[out_chan].cut(pos=(asp,asp+round(self.dur*self.fs))).delay(asp/self.fs)
            for i,s in enumerate(self.in_sig):
                self.in_sig[i] = s.cut(pos=(asp+ds+1,asp+ds+1+self.out_sig[out_chan].length))
                self.in_sig[i].t0 = self.out_sig[out_chan].t0
        return d

# def peak_sync_prepare(M,out_chan=0,in_chan=0):
#     def inner(*args,**kwargs):
#         osig = M.out_sig[out_chan]
#         osigpeak = osig.similar(
#             values=np.stack((picv(long=2*M.fs),osig.values)),
#             t0=osig.t0-2)
#         M1 = copy(M)
#         M1.out_sig[out_chan]=osigpeak
#         func(M1)
#         isig = M1.in_sig[in_chan]
#         posmax = int( np.argmax(isig.values[int(0.25*M.fs*2):int(0.75*M.fs*2)]) + 0.75*M.fs*2 )
#         print(posmax)
#         for i,s in enumerate(M1.in_sig):
#             s.values = s.values[posmax:posmax+M.fs*M.dur]
#         M=M1
#         del(M1)
#     return inner

    # def peak_sync_prepare(self,out_chan=0):
    #     osig = self.out_sig[out_chan]
    #     osigpeak = osig.similar(
    #         values=np.stack((picv(long=2*M.fs),osig.values)),
    #         t0=osig.t0-2)
    #     del(osig)
    #     M1 = deepcopy(self)
    #     M1.out_sig[out_chan]=osigpeak
    #     return M1

    # def peak_sync_render(self,in_chan=0):
    #     isig = self.in_sig[in_chan]
    #     posmax = int( np.argmax(isig.values[int(0.25*M.fs*2):int(0.75*M.fs*2)]) + 0.75*M.fs*2 )
    #     print(posmax)
    #     for i,s in enumerate(M1.in_sig):
    #         s.values = s.values[posmax:posmax+M.fs*M.dur]
    #     return M1
    #     del(M1)


# measpy/utils.py
#
# -----------------------------------
# Measpy : various utility functions
# -----------------------------------
#
# Part of measpy package for signal acquisition and processing
# (c) OD - 2021 - 2023
# https://github.com/odoare/measpy

import numpy as np
import scipy.io.wavfile as wav
from ._tools import siglist_to_array, all_equal
from .signal import PREF

def mic_calibration_level(sig, current_cal=1., target_db = 94.):
    """ Microphone calibration function

        :param sig: Signal recorded by the microphone to calibrate
        :type sig: measpy.signal.Signal
        :param current_cal: Calibration indicated in the recorded signal sig
        :type sig: float
        :param target_db: dB SPL level of the microphone calibrator
        :type target_db: float

        :return: The actual value to use as calibration level, so that measured dB SPL equals to target_db
        :rtype: float

        Howto use:
            - Record mic signal with pistonphone using

            .. highlight:: python
            .. code-block:: python

                import measpy as mp
                FS = 48000
                DUR = 10
                pa = mp.Signal(fs=FS, unit='Pa')
                M = mp.Measurement( device_type='audio',
                                    fs = FS,
                                    in_sig=[pa],
                                    in_map=[1],
                                    dur=DUR,
                                    in_device='Default')
                audio_run_measurement(M)
            
            - Recorded signal is then analysed to get actual calibration

            .. highlight:: python
            .. code-block:: python

                newcal = mic_calibration_level(M.in_sig[0],current_cal=1.)

            - Calibration is then the value to use at signal creation

    """
    measured_spl = 20*np.log10(sig.rms/PREF)
    return current_cal*10**( (measured_spl-target_db) / 20 )

def mic_calibration_freq(sig, sigref, Wref=None, noct=3, nperseg=None):
    """ Microphone frequency response calibration function

        :param sig: Signal recorded by the microphone to calibrate
        :type sig: measpy.signal.Signal
        :param sigref: Signal recorded by the reference microphone
        :type sigref: measpy.signal.Signal
        :param Wref: Reference microphone response function as a Weighting object
        :type Wref: measpy.signal.Weighting
        :param noct: Specifies the frequency averaging to apply. The returned measpy.signal.Weighting object will be 1/noct frequency averaged
        :type noct: int
        :param nperseg: Window size when transfer function estimation is made
        :type nperseg: int

        :return: A Weighting object containing the frequency response of the microphone
        :rtype: measpy.signal.Weighting

        Howto use:
            - Record mic and refmic signals using for example:

            .. highlight:: python
            .. code-block:: python

                import measpy as mp
                FS = 48000
                DUR = 10
                pa = mp.Signal(fs=FS, unit='Pa',cal=...)
                paref = mp.Signal(fs=FS, unit='Pa', cal=...)
                M = mp.Measurement( device_type='audio',
                                    fs = FS,
                                    in_sig=[pa,paref],
                                    in_map=[1,2],
                                    dur=DUR,
                                    in_device='Default')
                audio_run_measurement(M)

            - Recorded signal is then analysed to get actual calibration

            .. highlight:: python
            .. code-block:: python

                micresp = mic_calibration_freq(M.in_sig[0],M.in_sig[1],Wref=...)

            Wref=... corresponds to the reference mic response curve.

            - Calibration curve is then the returned measpy.Weighting object

    """
    if Wref==None:
        if nperseg==None:
            return abs(sig.tfe_welch(sigref)).nth_oct_smooth_to_weight(noct)
        else:
            return abs(sig.tfe_welch(sigref, nperseg)).nth_oct_smooth_to_weight(noct)
    else:
        if nperseg==None:
            return abs(sig.tfe_welch(sigref.rfft().apply_weighting(Wref,inverse=True).irfft())).nth_oct_smooth_to_weight(noct)
        else:
            return abs(sig.tfe_welch(sigref.apply_weighting(Wref,inverse=True), nperseg)).nth_oct_smooth_to_weight(noct)

def siglist_to_wav(sigl,filename):
    """ Takes a list of signals and export it to a multichannel wave file.
        Sampling frequencies have to match. Use the resample method if necessary.

        :param sigl: A list of single channel signals
        :type sigl: list of measpy.signal.Signal
        :param filename: Name of wave file (without extension)
        :type filename: string

        :return: True if succeded, False if error
        :rtype: bool

    """
    print([s.fs for s in sigl])
    if not(all_equal([s.fs for s in sigl])):
        print('Not possible: different sampling frequencies \n Resample first')
        return False
    wav.write(filename+'.wav', int(round(sigl[0].fs)), siglist_to_array(sigl))
    return True



# -*- coding: utf-8 -*-
"""
Created on Fri Mar 21 15:40:04 2025

@author: clement
"""
import numpy as np
from abc import ABC, abstractmethod
from queue import Empty

import matplotlib.pyplot as plt
import matplotlib as mpl

mpl.rcParams["path.simplify"] = True
mpl.rcParams["path.simplify_threshold"] = 1.0
mpl.rcParams["agg.path.chunksize"] = 10000

import time


def timing(f):
    def wrap(*args, **kwargs):
        time1 = time.time()
        ret = f(*args, **kwargs)
        time2 = time.time()
        print(
            "{:s} function took {:.3f} ms".format(
                f.__name__, (time2 - time1) * 1000.0
            )
        )

        return ret

    return wrap


def justify(a, axis=0, side="left"):
    """
    Justifies a 2D array
    :param a: Input array to be justified
    :type a: ndarray
    :param axis: Axis along which justification is to be made, defaults to 0
    :type axis: int, optional
    :param side: Direction of justification. It could be 'left', 'right', 'up', 'down'
    It should be 'left' or 'right' for axis=1 and 'up' or 'down' for axis=0., defaults to 'left'
    :type side: str, optional

    """

    mask = ~np.isnan(a)
    justified_mask = np.sort(mask, axis=axis)
    if (side == "up") | (side == "left"):
        justified_mask = np.flip(justified_mask, axis=axis)
    out = np.full(a.shape, np.nan)
    if axis == 1:
        out[justified_mask] = a[mask]
    else:
        out.T[justified_mask.T] = a.T[mask.T]
    return out


class BlitManager:
    def __init__(self, canvas, animated_artists=()):
        """
        Parameters
        ----------
        canvas : FigureCanvasAgg
            The canvas to work with, this only works for subclasses of the Agg
            canvas which have the `~FigureCanvasAgg.copy_from_bbox` and
            `~FigureCanvasAgg.restore_region` methods.

        animated_artists : Iterable[Artist]
            List of the artists to manage
        """
        self.canvas = canvas
        self._bg = None
        self._artists = []
        self.changed_axe = False

        for a in animated_artists:
            self.add_artist(a)
        # grab the background on every draw
        self.cid = canvas.mpl_connect("draw_event", self.on_draw)

    def on_draw(self, event):
        """Callback to register with 'draw_event'."""
        cv = self.canvas
        if event is not None:
            if event.canvas != cv:
                raise RuntimeError
        self._bg = cv.copy_from_bbox(cv.figure.bbox)
        self._draw_animated()

    def add_artist(self, art):
        """
        Add an artist to be managed.

        Parameters
        ----------
        art : Artist

            The artist to be added.  Will be set to 'animated' (just
            to be safe).  *art* must be in the figure associated with
            the canvas this class is managing.

        """
        if type(art) is list:
            for oneArt in art:
                if oneArt.figure != self.canvas.figure:
                    raise RuntimeError
                oneArt.set_animated(True)
                self._artists.append(oneArt)
        else:
            if art.figure != self.canvas.figure:
                raise RuntimeError
            art.set_animated(True)
            self._artists.append(art)

    def _draw_animated(self):
        """Draw all of the animated artists."""
        fig = self.canvas.figure
        for a in self._artists:
            fig.draw_artist(a)

    def update(self):
        """Update the screen with animated artists."""
        if self.changed_axe:
            self.canvas.draw()
            self.changed_axe = False
            plt.pause(0.0001)
        else:
            cv = self.canvas
            fig = cv.figure
            # paranoia in case we missed the draw event,
            if self._bg is None:
                self.on_draw(None)
            else:
                # restore the background
                cv.restore_region(self._bg)
                # draw all of the animated artists
                self._draw_animated()
                # update the GUI state
                cv.blit(fig.bbox)
            # let the GUI event loop process anything it has to do
            cv.flush_events()


class plot_data_from_queue(ABC):
    """
    Abstract class used to analyse and plot data that are feed into a queue (by a measurment callback)
    """

    plot_attribute = ["plotbuffer", "axes", "lines", "istimedata"]

    def __init__(
        self, fs, updatetime=0.1, plotbuffersize=2000, nchannel=1, show_time0=True
    ):
        """
        :param fs: Frequency (Hz) of the signal, used to define 'timeinterval' attribute
        that manage the time dependencies of the plot.
        :type fs: float
        :param updatetime: Time (second) between each update of the plot, defaults to 0.1
        :type updatetime: flaot, optional
        :param plotbuffersize: define attribute to store a size of the plot buffer
        that can be used inside 'plot_setup' method, defaults to 2000
        :type plotbuffersize: int, optional
        :param nchannel: Number of channels inside the Queue, defaults to 1
        :type nchannel: int, optional
        :param show_time0: If true, show a indicator for t0, defaults to True
        :type show_time0: bool, optional
        :return: Initialise the instance, with attribute useful to define plot and mandatory to use the class
        :rtype: plot_data_from_queue

        """
        for x in self.plot_attribute:
            setattr(self, x, None)
        self.timesincelastupdate = 0
        self.plotbuffersize = plotbuffersize
        self.updatetime = updatetime
        self.timeout = 0.1 * updatetime
        self.fs = fs
        self.timeinterval = 1 / self.fs
        self.plot_duration = plotbuffersize * self.timeinterval
        self.databuffersize = max(int(updatetime * self.fs), plotbuffersize)
        self.nchannel = nchannel
        if nchannel > 1:
            self.data_buffer = np.zeros((self.databuffersize, nchannel))
        else:
            self.data_buffer = np.zeros((self.databuffersize))
        animated_artists = self.plot_setup()
        if not animated_artists:
            animated_artists = []
        for x in self.plot_attribute + ["fig"]:
            if getattr(self, x) is None:
                raise TypeError(
                    f"Subclasses 'plot_setup' method must set {x} to a non-None value"
                )
        nlines = len(self.lines)
        for x in self.plot_attribute:
            if not len(getattr(self, x)) == nlines:
                raise ValueError(
                    f"The size of {x} is not the same as the number of lines = {nlines}"
                )

        animated_artists += self.lines
        self.tend = 0
        if show_time0:
            self.time0 = self.axes[0].text(
                0.05,
                0.05,
                "$t_0 = 0$",
                transform=self.axes[0].transAxes,
                va="bottom",
                ha="left",
                animated=True,
            )
            animated_artists += [self.time0]
        else:
            self.time0 = None
        self.bm = BlitManager(self.fig.canvas, animated_artists)

    @abstractmethod
    def plot_setup(self):
        """
        Create the plot and attributes used to modify the plot :
            - plotbuffer: list of Numpy array that contain data to be plotted
            - fig : matplotlib figure
            - axes : list of matplotlib axes
            - lines : list of  matplotlib lines
            - istimedata : list of boolean, if true the nan for the considered axis are displaced
            to the right (allow better display at begining)

        plotbuffer, axes, lines and istimedata should be lists of the same size,
        each element corresponds to one line of data plotted
        can return a animated artist to be updated inside data_process

        """
        pass

    @abstractmethod
    def data_process(self):
        """
        Process data to be plotted by updating plotbuffer using data_buffer
        data_buffer is a numpy array (dimensions = [databuffersize,nchannels])
        that contain the data from the Queue
        databuffersize is greater than plotbuffersize and updatetime * fs

        """
        pass

    def _plotting_buffer(self):
        self.data_process()
        for ax, line, data, istimedata in zip(
            self.axes, self.lines, self.plotbuffer, self.istimedata
        ):
            if istimedata:
                data = justify(data)
            line.set_ydata(data)
        self.rescaling()
        if self.time0 is not None:
            self.tend += self.timesincelastupdate * self.timeinterval
            self.time0.set_text(
                f"$t_0 = {max(0,self.tend-self.plot_duration):.2f}$"
            )
        self.bm.update()
        self.timesincelastupdate = 0

    def rescaling(self):
        """
        This method is called automatically to rescale the data after each plot
        By default, it does nothing.
        It has to set self.bm.change_axe to True when axes are changed

        """
        pass

    def _update_data_buffer(self, item):
        n_values = len(item)
        # item = np.asarray(item) * 0.001  #mv to V
        self.timesincelastupdate += n_values
        if n_values <= self.databuffersize:
            self.data_buffer[:-n_values] = self.data_buffer[n_values:]
            self.data_buffer[-n_values:] = item
        else:
            self.data_buffer[:] = item[-self.databuffersize :]

    def update_plot(self, updatetime=None):
        updatetime = self.updatetime if updatetime is None else updatetime
        try:
            if (item := self.dataqueue.get(timeout=self.timeout)) is not None:
                item = np.asarray(item).squeeze()
                self._update_data_buffer(item)
                if self.timesincelastupdate * self.timeinterval > updatetime:
                    self._plotting_buffer()
        except (Empty, AttributeError):
            pass

    def update_plot_until_empty(self):
        try:
            while (item := self.dataqueue.get(timeout=10)) is not None:
                item = np.asarray(item).squeeze()
                self._update_data_buffer(item)
                if self.timesincelastupdate * self.timeinterval > self.updatetime:
                    self._plotting_buffer()
            if self.timesincelastupdate > 0:
                self._plotting_buffer()
        except (Empty, AttributeError):
            pass

    def close(self):
        plt.close(self.fig)

    @property
    def dataqueue(self):
        try:
            return self._dataqueue
        except AttributeError:
            print("No dataqueue defined")
            return None

    @dataqueue.setter
    def dataqueue(self, dataqueue):
        if (item := dataqueue.get(timeout=100 * self.timeout)) is not None:
            item = np.asarray(item).squeeze()
            if item[0].size == self.data_buffer[0].size:
                self._update_data_buffer(item)
                if self.timesincelastupdate * self.timeinterval > self.updatetime:
                    self._plotting_buffer()
                self._dataqueue = dataqueue
            else:
                raise ValueError("Invalid queue")


# measpy/pico.py
#
# ---------------------------------------
# Data acquisition with Picoscope devices
# ---------------------------------------
#
# Part of measpy package for signal acquisition and processing
# (c) OD - 2021 - 2023
# https://github.com/odoare/measpy


import numpy as np
from inspect import getfullargspec

from datetime import datetime
from scipy.signal import decimate

import ctypes
from picosdk.ps4000 import ps4000
from picosdk.functions import assert_pico_ok, adc2mV, mV2adc

from threading import Thread, Event
from queue import Queue
import time
from functools import partial
from unyt import Unit
from unyt.exceptions import UnitConversionError

from picosdk.ps2000 import ps2000
from picosdk.functions import assert_pico2000_ok
from picosdk.ctypes_wrapper import C_CALLBACK_FUNCTION_FACTORY
from picosdk.PicoDeviceEnums import picoEnum


from abc import ABC, abstractmethod

from measpy._tools import H5file_valid

# plt.style.use('seaborn-v0_8')

maxtimeout = 10

Channel_state = {key: None for key in ["enabled", "coupling", "range", "buffer"]}

PS2000_channel = {"A": 1, "B": 2, 1: "A", 2: "B"}
PS4000_channel = {"A": 1, "B": 2, "C": 3, "D": 4, 1: "A", 2: "B", 3: "C", 4: "D"}


def transposelist(L):
    return [list(i) for i in zip(*L)]


def dispatch(q_in, qs_out):
    while (item := q_in.get(timeout=maxtimeout)) is not None:
        for q_out in qs_out:
            q_out.put(item)
    for q_out in qs_out:
        q_out.put(None)


def convert_to_mutichannel(listSig):
    ret = listSig[0]
    for i in range(len(listSig) - 1):
        ret = ret.pack_with(listSig[i + 1])
    return ret


def Fill_Signal_Queue(Sig, q_in, unit_in="mV", Ndata=None):
    for s in Sig:
        if s.unit == Unit("1"):
            s.unit = "mV"
            conversion = 1.0
        else:
            try:
                conversion = (1.0 * Unit("mV")).to_value(s.unit)
            except UnitConversionError:
                print(
                    f"Signal unit ({s.unit}) incompatible with mV, unit set to mV"
                )
                s.unit = "mV"
                conversion = 1.0
    if Ndata:
        array = np.zeros((Ndata, Sig.nchannels)).squeeze()
        Sig.raw = conversion * Queue2prealocated_array(q_in, array)
    else:
        Sig.raw = conversion * Queue2array(q_in)


def Queue2prealocated_array(q_in, array):
    datasize = array.shape[0]
    nextSample = 0
    while (item := q_in.get(timeout=maxtimeout)) is not None:
        item = np.asarray(item).transpose()
        noOfSamples = item.shape[0]
        lastsample = nextSample + noOfSamples
        try:
            array[nextSample:lastsample] = item
            nextSample = lastsample
        except ValueError:
            N = datasize - nextSample
            if N > 0:
                array[nextSample:] = item[:N]
            break
    return array


def Queue2array(q_in):
    data = q_in.get(timeout=maxtimeout)
    while (item := q_in.get(timeout=maxtimeout)) is not None:
        item = np.asarray(item).transpose()
        data =  np.concatenate([data,item],axis = 1)
    return data


def findindex(l, e):
    try:
        a = l.index(e)
    except:
        a = None
    return a


def detect_rising_pulses_threshold_ind(
    values,
    threshold,
    ind0,
    previous_data_point,
):
    """
    Detect rising pulse using a threshold

    :param values: Data.
    :type values: List
    :param ind0: indice of the first data point.
    :type ind0: int
    :param previous_data_point: Value of the last data point (indice = ind0-1).
        useful to not lose the peak at ind0
    :type previous_data_point: list with one number
    :param threshold: threshold (in adc values).
    :type threshold: list with one int
    :return: List of indices where a rising pulse is detected.
    :rtype: List
    """
    V = np.asarray(previous_data_point + values)
    rising = np.flatnonzero((V[:-1] <= threshold) & (V[1:] > threshold))
    previous_data_point = [values[-1]]
    id0 = ind0[0]
    ind0[0] += len(values)
    return rising + id0


def ps2000_run_measurement(M, serial=None, filename=None, **kwargs):
    with _ps2000_run_measurement_threaded(
        M, filename=filename, serial=serial, **kwargs
    ) as device:
        device.start()
        device.join()


def ps2000_plot(
    M, plotting_class, plotbuffersize=20000, updatetime=0.1, serial=None, **kwargs
):
    queueplot = Queue()
    nchannel = len(M.in_sig)
    plotting_instance = plotting_class(
        M.fs,
        updatetime=updatetime,
        plotbuffersize=plotbuffersize,
        nchannel=nchannel,
    )
    with _ps2000_run_measurement_threaded(
        M,
        output_queue=queueplot,
        serial=serial,
        stop_event=plotting_instance.stop_event,
        **kwargs,
    ) as device:
        device.start()
        plotting_instance.dataqueue = queueplot
        plotting_instance.update_plot_until_empty()
        device.join()


def ps2000_pulse_detection(
    M, serial=None, min_chunksize_processed=10000, **kwargs
):
    """
    This function needs M to contain the property
     in_threshold : a list of threshold for pulse height (in Volt) for each channel
    """
    if hasattr(M, "in_threshold") and M.in_threshold is not None:
        Q = Queue()
        ps2000_run_measurement(
            M,
            pre_process=detect_rising_pulses_threshold_ind,
            save_into_Signal=False,
            output_queue=Q,
            min_chunksize_processed=min_chunksize_processed,
            serial=serial,
            **kwargs,
        )
        return np.double(Queue2array(Q)) / M.fs
    else:
        raise ValueError("There is no threshold defined")


def ps4000_run_measurement(M, serial=None, filename=None, **kwargs):
    with _ps4000_run_measurement_threaded(
        M, filename=filename, serial=serial, **kwargs
    ) as device:
        device.start()
        device.join()


def ps4000_plot(
    M, plotting_class, plotbuffersize=20000, updatetime=0.1, serial=None, **kwargs
):
    queueplot = Queue()
    nchannel = len(M.in_sig)
    plotting_instance = plotting_class(
        M.fs,
        updatetime=updatetime,
        plotbuffersize=plotbuffersize,
        nchannel=nchannel,
    )
    with _ps4000_run_measurement_threaded(
        M,
        output_queue=queueplot,
        serial=serial,
        stop_event=plotting_instance.stop_event,
        **kwargs,
    ) as device:
        device.start()
        plotting_instance.dataqueue = queueplot
        plotting_instance.update_plot_until_empty()
        device.join()


def ps4000_pulse_detection(
    M, serial=None, min_chunksize_processed=10000, **kwargs
):
    """
    This function needs M to contain the property
     in_threshold : a list of threshold for pulse height (in Volt) for each channel
    """
    if hasattr(M, "in_threshold") and M.in_threshold is not None:
        Q = Queue()
        ps4000_run_measurement(
            M,
            pre_process=detect_rising_pulses_threshold_ind,
            save_into_Signal=False,
            output_queue=Q,
            min_chunksize_processed=min_chunksize_processed,
            serial=serial,
            **kwargs,
        )
        return np.double(Queue2array(Q)) / M.fs
    else:
        raise ValueError("There is no threshold defined")


class Pico_thread(ABC, Thread):
    """
    This needs M to contain the following properties:
        - in_range : a list of strings specifying the voltage range.
            Possible voltage ranges are "10MV", "20MV", "50MV", "100MV",
            "200MV", "500MV", "1V", "2V", "5V", "10V", "20V", "50V", "100V"
        - upsampling_factor : upsampling factor
        - in_coupling : Coupling configuration of the channels.
            Can be "ac" or "dc"
    """

    dataqueue = Queue()
    Channels = {}
    channel_name = {}
    Data_type = {}

    def __init__(
        self,
        M,
        output_queue=None,
        pre_process=adc2mV,
        save_into_Signal=True,
        min_chunksize_processed=0,
        filename=None,
        buffersize=20_000,
        max_samples=100_000,
        serial=None,
        stop_event=Event(),
    ):
        """
        :param M: Parameter container
        :type M: measpy.measurement.
        :param output_queue: Send data into this queue, defaults to None
        :type output_queue: queue.Queue
        :param pre_process: Method to process raw data, defaults to adc2mV
        :type pre_process: callable
        :param save_into_Signal: Saving the data into the signal, defaults to True
        :type save_into_Signal: bool, optional
        :param min_chunksize_processed: Minimum size of data to process at once, defaults to 0
        :type min_chunksize_processed: int, optional
        :param filename: Hdf5 file path to direct save into disk, defaults to None
        :type filename: str or pathlib.Path, optional
        :param buffersize: Size of picoscope buffer, defaults to 20_000
        :type buffersize: int, optional
        :param max_samples: Maximum number of sample that stay in memory, defaults to 100_000
        :type max_samples: int, optional
        :param serial: Serial number of the scope, defaults to None
        :type serial: str, optional
        :param stop_event: Event that stop measurment when set, can be defined or used
        :type stop_event: threading.Event()
        """
        # init threading, deamon = True to keyboard interupt when threading
        super().__init__(daemon=True)
        # easier to use only multichannel signal
        if isinstance(M.in_sig, list):
            M.in_sig = convert_to_mutichannel(M.in_sig)
        self.M = M
        if M.device_type != "pico":
            raise ValueError("Error: device_type must be 'pico'.")

        if type(M.out_sig) != type(None):
            print("Warning: out_sig property ignored with picoscopes")

        ## parameters
        self.stop_measurement = stop_event
        self.min_chunksize_processed = min_chunksize_processed
        self.pre_process = pre_process

        ## Buffer are always read in the same order, remapping to match in_map
        self.buffers_map = np.argsort(np.argsort(self.M.in_map))

        # serial number of the scope
        if serial is None:
            print("No picoscope given, searching for connected picoscope")
        self.serial = serial

        # Number of sample stored by picoscope driver that can be retrieved after
        # streaming stopped (by ps2000_get_streaming_values or ps4000GetValuesAsyn)
        self.max_samples = max_samples
        # Size of the temporary buffers used for storing the data
        self.overview_buffer_size = buffersize
        self.save_into_Signal = save_into_Signal
        self.output_queue = output_queue

        # If filename given, save data in hdf5 file
        self.filename = filename
        self.savehdf5 = False
        if H5file_valid(self.filename):
            print(f"Raw measurment will be save in {filename}")
            self.savehdf5 = True

        # Effective sampling frequency
        # If upsampling_factor is > 1, the actual data acquisition is
        # performed at fs*upsampling_factor, and then decimated to
        # the desired frequency

        self.effective_fs = M.fs * M.upsampling_factor
        self.duree_ns = M.dur * 1e9

        # Sample interval is the duration between two consecutive samples
        # As it is the sampling frequency that is specified,
        # and the sampling interval can only take integer increments
        # of 1 nanoseconds, the actual sampling frequency might
        # necessitate adjustments

        if (si := (1e9 / self.effective_fs)).is_integer():
            self.sampleInterval = ctypes.c_int32(int(si))
        else:
            si = int(round(si))
            self.sampleInterval = ctypes.c_int32(si)
            self.effective_fs = 1e9 / si
            print(
                "Warning : Effective sampling frequency ",
                f"changed to nearest possible value of {self.effective_fs} Hz",
            )
            M.fs = self.effective_fs / M.upsampling_factor
            for s in M.in_sig:
                s.fs = M.fs
            print(
                "Warning : Signal sampling frequency fs ",
                f"changed to nearest possible value of {M.fs} Hz",
            )

        numdesiredsamples = int(round(self.effective_fs * self.M.dur))
        numBuffersToCapture = int(
            np.ceil(numdesiredsamples / self.overview_buffer_size)
        )
        self.totalSamples = self.overview_buffer_size * numBuffersToCapture
        self.time_to_fill_half_buffer = (
            0.5 * self.overview_buffer_size / self.M.fs
        )

    def __enter__(self):
        # open comunication with picoscope and set up channel and preprocesing tasks
        self.open_unit()
        try:
            self.channels_setup()
            self.process = self.setup_threads()
        except Exception as e:
            self.__exit__(type(e), e.args, e)
            raise e
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # Close cominucation
        self.close_unit()

    def channels_setup(self):
        # Use Measurment data to set up the channels
        for chan in self.Channels.values():
            i = chan["number"]
            self.setup_channel(i)
            if (
                (Vthreshold := getattr(self.M, "in_threshold", None)) is not None
            ) and (ind := findindex(self.M.in_map, i)) != None:
                chan["adc_threshold"] = mV2adc(
                    [Vthreshold[ind] * 0.001],
                    chan["range"],
                    self.Data_type["max"],
                )[0]
            else:
                chan["adc_threshold"] = None
        self.nchannels = sum([chan["enabled"] for chan in self.Channels.values()])

    def run(self):
        """
        self.start() run this into a thread

        """
        ##start preprocess threads
        for P in self.process:
            P.start()
        # run measurment
        self.start_measurment()

        # Wait for all threads to finish
        print("Waiting for data processing...")
        for P in self.process:
            P.join()

        # Upsampling
        if self.save_into_Signal and self.M.upsampling_factor > 1:
            values = self.M.in_sig.raw
            self.M.in_sig.raw = decimate(
                np.double(values), self.M.upsampling_factor, ftype="fir"
            )
        print("Preprocess data done")

    def create_h5file(self):
        """
        save h5file with dbfs according to channel parameter
        :return: None
        :rtype: None

        """
        dbfs = []
        for sig, prange in zip(self.M.in_sig, self.M.in_range):
            sig.unit = "mV"
            dbfs.append(adc2mV([1], self.pico_range(prange), self.Data_type)[0])
        self.M.datatype = np.dtype(self.Data_type).name
        print("Creating the H5file with measurment parameters")
        self.h5save_data = self.M.create_hdf5(self.filename, dbfs = dbfs)

    def setup_threads(self):
        # threads and outputs set up (preprocessing, put into queue, put into Signal)
        ret = []
        if self.savehdf5:
            # In hdf5 ADC data (integers) are directly saved to file
            # dbfs convert ADC to mVolt and is saved in the parameters
            self.create_h5file()
            queuesave = Queue()
            queueprocess = Queue()
            ret.append(
                Thread(
                    target=dispatch,
                    args=(self.dataqueue, [queuesave, queueprocess]),
                )
            )
            ret.append(Thread(target=self.h5save_data, args=(queuesave,)))
        else:
            queueprocess = self.dataqueue
        process_data = self.setup_preprocess()
        if self.save_into_Signal:
            processed_queue = Queue()
            if self.output_queue:
                queueSignal = Queue()
                ret.append(
                    Thread(
                        target=process_data, args=(queueprocess, processed_queue)
                    )
                )
                ret.append(
                    Thread(
                        target=dispatch,
                        args=(processed_queue, [self.output_queue, queueSignal]),
                    )
                )
                ret.append(
                    Thread(
                        target=Fill_Signal_Queue,
                        args=(self.M.in_sig, queueSignal),
                        kwargs={
                            "Ndata": int(
                                self.M.dur * self.M.fs * self.M.upsampling_factor
                            )
                        },
                    )
                )
            else:
                ret.append(
                    Thread(
                        target=process_data, args=(queueprocess, processed_queue)
                    )
                )
                ret.append(
                    Thread(
                        target=Fill_Signal_Queue,
                        args=(self.M.in_sig, processed_queue),
                        kwargs={
                            "Ndata": int(
                                self.M.dur * self.M.fs * self.M.upsampling_factor
                            )
                        },
                    )
                )
        elif self.output_queue:
            ret.append(
                Thread(
                    target=process_data, args=(queueprocess, self.output_queue)
                )
            )
        elif not self.savehdf5:
            raise ValueError("There are not output for data")
        return ret

    def methods_setup(self):
        # use data in measurment and from channel setup to "partial" the preprocessing method (convert to mV or other)
        if callable(self.pre_process):
            self.pre_process = [self.pre_process] * self.nchannels
        methods = []
        for method, ind in zip(self.pre_process, self.M.in_map):
            chan = self.Channels[self.channel_name[ind]]
            method_args = getfullargspec(method).args
            # range, threshold and maxADC: optional argument not used by the func method
            kwargs = {
                "range": chan["range"],
                "threshold": chan["adc_threshold"],
                "maxADC": self.Data_type,
                "ind0": [0],
                "previous_data_point": [],
            }
            partial_kwargs = {a: b for a, b in kwargs.items() if a in method_args}
            if partial_kwargs:
                methods.append(partial(method, **partial_kwargs))
            else:
                methods.append(method)

        def process(values):
            return [method(values[ind]) for (ind, method) in zip(self.buffers_map, methods)]

        return process

    def setup_preprocess(self):
        # define func that preprocess raw data
        preprocess = self.methods_setup()

        if self.min_chunksize_processed > 0:

            def func(queue, queueout):
                chunk_data = [[] for _ in range(self.nchannels)]
                while (item := queue.get(timeout=maxtimeout)) is not None:
                    _ = [c.extend(it) for (c, it) in zip(chunk_data, item)]
                    if len(chunk_data[0]) > self.min_chunksize_processed:
                        values = preprocess(chunk_data)
                        queueout.put(values)
                        chunk_data = [[] for _ in range(self.nchannels)]
                if len(chunk_data[0]) > 0:
                    values = preprocess(chunk_data)
                    queueout.put(values)
                queueout.put(None)

        else:

            def func(queue, queueout):
                while (item := queue.get(timeout=maxtimeout)) is not None:
                    values = preprocess(item)
                    queueout.put(values)
                queueout.put(None)

        return func

    @abstractmethod
    def open_unit(self):
        """
        Open picoscope

        """
        pass

    @abstractmethod
    def close_unit(self):
        """
        Close picoscope

        """
        pass

    @abstractmethod
    def setup_channel(self, chan_index):
        """setup a channel.

        should define channel status (used or not), channel coupling(ac or dc),
        channel range and channel data buffer.
        """

    @abstractmethod
    def setup_callback(self):
        """
        Setup the callback method to read data.
        should return the callback metod used to read and put data into dataqueue

        """
        pass

    @abstractmethod
    def start_measurment(self):
        """
        Run the scope.
        """

    @staticmethod
    @abstractmethod
    def pico_range(prange):
        """
        Picoscope voltage range (mapped from picoscope lib)
        This property should be overridden, used by create_h5file and setup_channel
        """
        pass


class _ps2000_run_measurement_threaded(Pico_thread):
    Channels = {
        "A": {"number": 1} | Channel_state.copy(),
        "B": {"number": 2} | Channel_state.copy(),
    }
    channel_name = {1: "A", 2: "B"}
    Data_type = ctypes.c_int16(32767)

    def open_unit(self):
        self.device = ps2000.open_unit(serial=self.serial)
        print("Device info: {}".format(self.device.info))
        self.closed = False

    def close_unit(self):
        if not self.closed:
            # stop
            ps2000.ps2000_stop(self.device.handle)
            # disconect
            self.device.close()
            self.closed = True

    def setup_channel(self, chan_index):
        if (ind := findindex(self.M.in_map, chan_index)) != None:
            enabled = True
            pico_range = self.pico_range(self.M.in_range[ind])
            if self.M.in_coupling[ind].capitalize() == "Dc":
                coupling = "PICO_DC"
            elif self.M.in_coupling[ind].capitalize() == "Ac":
                coupling = "PICO_AC"
            else:
                print(
                    f"Input {self.channel_name[chan_index]} coupling not recognized, set to 'dc'"
                )
                self.M.in_coupling[ind] = "dc"
                coupling = "PICO_DC"
            res = ps2000.ps2000_set_channel(
                self.device.handle,
                picoEnum.PICO_CHANNEL[
                    f"PICO_CHANNEL_{self.channel_name[chan_index]}"
                ],
                enabled,
                picoEnum.PICO_COUPLING[coupling],
                pico_range,
            )
            assert_pico2000_ok(res)
            print(
                f"Channel {self.channel_name[chan_index]}: enabled with range "
                + "PS2000_"
                + self.M.in_range[ind]
                + " ("
                + str(pico_range)
                + ")"
            )
        else:
            enabled = False
            pico_range = self.pico_range("10V")
            coupling = "PICO_DC"
            print(f"Channel {self.channel_name[chan_index]}: disabled")

        self.Channels[self.channel_name[chan_index]]["enabled"] = enabled
        self.Channels[self.channel_name[chan_index]]["range"] = pico_range
        self.Channels[self.channel_name[chan_index]]["coupling"] = coupling

    def setup_callback(self):
        CALLBACK = C_CALLBACK_FUNCTION_FACTORY(
            None,
            ctypes.POINTER(ctypes.POINTER(ctypes.c_int16)),
            ctypes.c_int16,
            ctypes.c_uint32,
            ctypes.c_int16,
            ctypes.c_int16,
            ctypes.c_uint32,
        )
        ##Define get_overview_buffers for channel configuration
        if self.nchannels == 2:

            def get_overview_buffers(
                buffers,
                _overflow,
                _triggered_at,
                _triggered,
                _auto_stop,
                n_values,
            ):
                self.dataqueue.put(
                    [buffers[0][0:n_values].copy(), buffers[2][0:n_values].copy()]
                )

        elif self.Channels["A"]["enabled"]:

            def get_overview_buffers(
                buffers,
                _overflow,
                _triggered_at,
                _triggered,
                _auto_stop,
                n_values,
            ):
                self.dataqueue.put([buffers[0][0:n_values].copy()])

        elif self.Channels["B"]["enabled"]:

            def get_overview_buffers(
                buffers,
                _overflow,
                _triggered_at,
                _triggered,
                _auto_stop,
                n_values,
            ):
                self.dataqueue.put([buffers[2][0:n_values].copy()])

        return CALLBACK(get_overview_buffers)

    def start_measurment(self):
        callback = self.setup_callback()
        ########## Signal generator ##########
        if self.M.sig_gen:
            res = ps2000.ps2000_set_sig_gen_built_in(
                self.device.handle,
                int(self.M.offset * 1e6),  # offset voltage in uV
                int(self.M.amp / 2 * 1e6),  # peak-to-peak votage in uV
                self.M.wave,  # type of waveform (0 = sine wave)
                self.M.freq_start,  # start frequency in Hz
                self.M.freq_stop,  # stop frequency in Hz
                self.M.freq_change,  # frequency change per interval in Hz
                self.M.freq_int,  # interval of frequency change in seconds
                self.M.sweep_dir,  # sweep direction (0 = up)
                self.M.sweep_number,  # number of times to sweep
            )
            assert_pico2000_ok(res)
        #######################################
        # Time unit = nanosecond
        sampleUnits = 2
        # no autostop
        autoStopOn = False
        # No downsampling:
        downsampleRatio = 1

        res = ps2000.ps2000_run_streaming_ns(
            self.device.handle,
            self.sampleInterval.value,
            sampleUnits,
            self.max_samples,
            autoStopOn,
            downsampleRatio,
            self.overview_buffer_size,
        )
        assert_pico2000_ok(res)

        now = datetime.now()
        self.M.date = now.strftime("%Y-%m-%d")
        self.M.time = now.strftime("%H:%M:%S")

        print("Start measurement")
        start_time = time.time_ns()

        # loop until wanted duration + maximum time of one loop without buffer overrun or until stop_measurement is set
        margin = min(int(self.max_loop_time * 1e9), self.duree_ns * 0.1)
        try:
            while (
                time.time_ns() - start_time < self.duree_ns + margin
                and not self.stop_measurement.is_set()
            ):
                ps2000.ps2000_get_streaming_last_values(
                    self.device.handle, callback
                )
        except KeyboardInterrupt:
            pass
        self.dataqueue.put(None)
        print("Measurment done")
        overrun = False
        ps2000.ps2000_overview_buffer_status(self.device.handle, overrun)
        if overrun:
            print("Buffer have overrun")
        # Stop the scope
        self.close_unit()

    @staticmethod
    def pico_range(prange):
        return ps2000.PS2000_VOLTAGE_RANGE["PS2000_" + prange]


class _ps4000_run_measurement_threaded(Pico_thread):
    Channels = {
        "A": {"number": 1} | Channel_state.copy(),
        "B": {"number": 2} | Channel_state.copy(),
    }
    channel_name = {1: "A", 2: "B"}
    Data_type = ctypes.c_int16(32767)

    def open_unit(self):
        # picoscope 4000 handle
        self.chandle = ctypes.c_int16()
        # picoscope 4000 status
        self.status = {}
        # Open PicoScope 4000 Series device
        # Returns handle to chandle for use in future API functions
        if self.serial is not None:
            self.status["openunit"] = ps4000.ps4000OpenUnitEx(
                ctypes.byref(self.chandle),
                ctypes.c_char_p(bytes(self.serial, "ascii")),
            )
        else:
            self.status["openunit"] = ps4000.ps4000OpenUnit(
                ctypes.byref(self.chandle),
            )
        assert_pico_ok(self.status["openunit"])
        self.closed = False

    def close_unit(self):
        if not self.closed:
            # Stop the scope
            self.status["stop"] = ps4000.ps4000Stop(self.chandle)
            assert_pico_ok(self.status["stop"])

            # Disconnect the scope
            # handle = chandle
            self.status["close"] = ps4000.ps4000CloseUnit(self.chandle)
            assert_pico_ok(self.status["close"])
            self.closed = True

    def setup_channel(self, chan_index):
        if (ind := findindex(self.M.in_map, chan_index)) != None:
            enabled = True
            pico_range = self.pico_range(self.M.in_range[ind])
            if self.M.in_coupling[ind].capitalize() == "Dc":
                coupling = 1
            elif self.M.in_coupling[ind].capitalize() == "Ac":
                coupling = 0
            else:
                print(
                    f"Input {self.channel_name[chan_index]} coupling not recognized, set to 'dc'"
                )
                self.M.in_coupling[ind] = "dc"
                coupling = 1
            # Configure picoscope channel
            self.status[
                f"setCh{self.channel_name[chan_index]}"
            ] = ps4000.ps4000SetChannel(
                self.chandle,
                ps4000.PS4000_CHANNEL[
                    f"PS4000_CHANNEL_{self.channel_name[chan_index]}"
                ],
                int(enabled),
                coupling,
                pico_range,
            )
            assert_pico_ok(self.status[f"setCh{self.channel_name[chan_index]}"])
            print(
                f"Channel {self.channel_name[chan_index]}: enabled with range "
                + "PS4000_"
                + self.M.in_range[ind]
                + " ("
                + str(pico_range)
                + ")"
            )
            # Create buffer for data collection
            buffer = np.zeros(shape=self.overview_buffer_size, dtype=np.int16)
            # Assigning pointer's buffer
            self.status[
                f"setDataBuffers{self.channel_name[chan_index]}"
            ] = ps4000.ps4000SetDataBuffers(
                self.chandle,
                ps4000.PS4000_CHANNEL[
                    f"PS4000_CHANNEL_{self.channel_name[chan_index]}"
                ],
                buffer.ctypes.data_as(ctypes.POINTER(ctypes.c_int16)),
                None,
                self.overview_buffer_size,
            )
            assert_pico_ok(
                self.status[f"setDataBuffers{self.channel_name[chan_index]}"]
            )
        else:
            enabled = False
            pico_range = self.pico_range("10V")
            coupling = 1
            print(f"Channel {self.channel_name[chan_index]}: disabled")
            buffer = None
        self.Channels[self.channel_name[chan_index]]["enabled"] = enabled
        self.Channels[self.channel_name[chan_index]]["range"] = pico_range
        self.Channels[self.channel_name[chan_index]]["coupling"] = coupling
        self.Channels[self.channel_name[chan_index]]["buffer"] = buffer

    def setup_callback(self):
        """
        Create threads queues and callback metod

        :param chan_to_plot: Channel to plot
        :type chan_to_plot: str
        :raises NotImplementedError: save in hdf5 should be multichannel
        :return: calback function used by picoscope
        :rtype: method

        """
        ##For each case setup process and define get_overview_buffers
        self.nextSample = 0
        self.autoStopOuter = False
        self.wasCalledBack = False
        # listbuff = [chan["buffer"] for chan in self.Channels.values() if chan["enabled"]]
        # def streaming_callback(
        #     handle,
        #     noOfSamples,
        #     startIndex,
        #     overflow,
        #     triggerAt,
        #     triggered,
        #     autoStop,
        #     param,
        # ):
        #     self.wasCalledBack = True
        #     sourceEnd = startIndex + noOfSamples
        #     self.dataqueue.put([buff[startIndex:sourceEnd].copy() for buff in listbuff])
        #     if autoStop:
        #         self.autoStopOuter = True

        if self.nchannels == 2:

            def streaming_callback(
                handle,
                noOfSamples,
                startIndex,
                overflow,
                triggerAt,
                triggered,
                autoStop,
                param,
            ):
                self.wasCalledBack = True
                sourceEnd = startIndex + noOfSamples
                self.dataqueue.put(
                    [
                        self.Channels["A"]["buffer"][startIndex:sourceEnd].copy(),
                        self.Channels["B"]["buffer"][startIndex:sourceEnd].copy(),
                    ]
                )
                if autoStop:
                    self.autoStopOuter = True

        elif self.Channels["A"]["enabled"]:

            def streaming_callback(
                handle,
                noOfSamples,
                startIndex,
                overflow,
                triggerAt,
                triggered,
                autoStop,
                param,
            ):
                self.wasCalledBack = True
                sourceEnd = startIndex + noOfSamples
                self.dataqueue.put(
                    [self.Channels["A"]["buffer"][startIndex:sourceEnd].copy()]
                )
                self.nextSample += noOfSamples
                if autoStop:
                    self.autoStopOuter = True

        elif self.Channels["B"]["enabled"]:

            def streaming_callback(
                handle,
                noOfSamples,
                startIndex,
                overflow,
                triggerAt,
                triggered,
                autoStop,
                param,
            ):
                self.wasCalledBack = True
                sourceEnd = startIndex + noOfSamples
                self.dataqueue.put(
                    self.Channels["B"]["buffer"][startIndex:sourceEnd].copy()
                )
                self.nextSample += noOfSamples
                if autoStop:
                    self.autoStopOuter = True

        return ps4000.StreamingReadyType(streaming_callback)

    def start_measurment(self):
        callback = self.setup_callback()
        now = datetime.now()
        self.M.date = now.strftime("%Y-%m-%d")
        self.M.time = now.strftime("%H:%M:%S")

        # Begin streaming mode:
        sampleUnits = ps4000.PS4000_TIME_UNITS["PS4000_NS"]
        # We are not triggering:
        maxPreTriggerSamples = 0
        autoStopOn = 1
        # No downsampling:
        downsampleRatio = 1
        self.status["runStreaming"] = ps4000.ps4000RunStreaming(
            self.chandle,
            ctypes.byref(self.sampleInterval),
            sampleUnits,
            maxPreTriggerSamples,
            self.totalSamples,
            autoStopOn,
            downsampleRatio,
            self.overview_buffer_size,
        )
        assert_pico_ok(self.status["runStreaming"])

        print(f"Capturing at sample interval {self.sampleInterval.value} ns")
        try:
            while (
                self.nextSample < self.totalSamples
                and not self.autoStopOuter
                and not self.stop_measurement.is_set()
            ):
                self.wasCalledBack = False
                self.status[
                    "getStreamingLastestValues"
                ] = ps4000.ps4000GetStreamingLatestValues(
                    self.chandle, callback, None
                )
                if not self.wasCalledBack:
                    # If we weren't called back by the driver, this means no data is ready. Sleep for a short while before trying
                    # again.
                    time.sleep(self.time_to_fill_half_buffer)
        except KeyboardInterrupt:
            pass
        self.dataqueue.put(None)
        print("Measurment done")
        # Stop the scope
        self.close_unit()

    @staticmethod
    def pico_range(prange):
        return ps4000.PS4000_RANGE["PS4000_" + prange]
